package envoy.client.database;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

import envoy.client.ConfigItem;
import envoy.data.IdGenerator;
import envoy.util.SerializationUtils;

/**
 * Implements a {@link LocalDb} in a way that stores all information inside a
 * folder on the local file system.<br>
 * <br>
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>PersistentLocalDb.java</strong><br>
 * Created: <strong>27.10.2019</strong><br>
 *
 * @author Kai S. K. Engelbart
 * @author Maximilian K&auml;fer
 * @since Envoy v0.1-alpha
 */
public class PersistentLocalDb extends LocalDb {

	private File localDBDir, localDBFile, usersFile, idGeneratorFile;

	/**
	 * Initializes an empty local database without a directory. All changes made to
	 * this instance cannot be saved to the file system.<br>
	 * <br>
	 * This constructor shall be used in conjunction with the {@code ignoreLocalDB}
	 * {@link ConfigItem}.
	 *
	 * @since Envoy v0.3-alpha
	 */
	public PersistentLocalDb() {}

	/**
	 * Constructs an empty local database. To serialize any chats to the file
	 * system, call {@link PersistentLocalDb#initializeUserStorage()}.
	 *
	 * @param localDBDir the directory in which to store users and chats
	 * @throws IOException if the PersistentLocalDb could not be initialized
	 * @since Envoy v0.1-alpha
	 */
	public PersistentLocalDb(File localDBDir) throws IOException {
		this.localDBDir = localDBDir;

		// Initialize local database directory
		if (localDBDir.exists() && !localDBDir.isDirectory())
			throw new IOException(String.format("LocalDBDir '%s' is not a directory!", localDBDir.getAbsolutePath()));
		usersFile		= new File(localDBDir, "users.db");
		idGeneratorFile	= new File(localDBDir, "id_generator.db");
	}

	/**
	 * Creates a database file for a user-specific list of chats.
	 *
	 * @throws NullPointerException if the client user is not yet specified
	 * @since Envoy v0.1-alpha
	 */
	@Override
	public void initializeUserStorage() {
		if (user == null) throw new NullPointerException("Client user is null");
		localDBFile = new File(localDBDir, user.getId() + ".db");
	}

	@Override
	public void save() throws IOException {
		// Save users
		SerializationUtils.write(usersFile, users);

		// Save chats
		if (user != null) SerializationUtils.write(localDBFile, chats);

		// Save id generator
		if (hasIdGenerator()) SerializationUtils.write(idGeneratorFile, idGenerator);
	}

	@Override
	public void loadUsers() throws ClassNotFoundException, IOException { users = SerializationUtils.read(usersFile, HashMap.class); }

	@Override
	public void loadChats() throws ClassNotFoundException, IOException { chats = SerializationUtils.read(localDBFile, ArrayList.class); }

	@Override
	public void loadIdGenerator() {
		try {
			idGenerator = SerializationUtils.read(idGeneratorFile, IdGenerator.class);
		} catch (ClassNotFoundException | IOException e) {}
	}
}