package envoy.client.ui;

import java.awt.*;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.*;
import javax.swing.border.EmptyBorder;

import envoy.client.Chat;
import envoy.client.Settings;
import envoy.client.database.LocalDb;
import envoy.client.event.MessageCreationEvent;
import envoy.client.event.ThemeChangeEvent;
import envoy.client.net.Client;
import envoy.client.ui.list.ComponentList;
import envoy.client.ui.settings.SettingsScreen;
import envoy.client.util.EnvoyLog;
import envoy.data.Message;
import envoy.data.Message.MessageStatus;
import envoy.data.MessageBuilder;
import envoy.data.User;
import envoy.event.EventBus;
import envoy.event.MessageStatusChangeEvent;

/**
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>ChatWindow.java</strong><br>
 * Created: <strong>28 Sep 2019</strong><br>
 *
 * @author Kai S. K. Engelbart
 * @author Maximilian K&auml;fer
 * @author Leon Hofmeister
 * @since Envoy v0.1-alpha
 */
public class ChatWindow extends JFrame {

	// User specific objects
	private Client	client;
	private LocalDb	localDb;

	// GUI components
	private JPanel					contentPane				= new JPanel();
	private PrimaryTextArea			messageEnterTextArea	= new PrimaryTextArea(space);
	private JList<User>				userList				= new JList<>();
	private Chat					currentChat;
	private ComponentList<Message>	messageList				= new ComponentList<>(new MessageListRenderer());
	private PrimaryScrollPane		scrollPane				= new PrimaryScrollPane();
	private JTextPane				textPane				= new JTextPane();
	private PrimaryButton			postButton				= new PrimaryButton("Post");
	private PrimaryButton			settingsButton			= new PrimaryButton("Settings");

	private static final Logger logger = EnvoyLog.getLogger(ChatWindow.class.getSimpleName());

	// GUI component spacing
	private final static int	space	= 4;
	private static final Insets	insets	= new Insets(space, space, space, space);

	private static final long serialVersionUID = 6865098428255463649L;

	/**
	 * Initializes a {@link JFrame} with UI elements used to send and read messages
	 * to different users.
	 *
	 * @since Envoy v0.1-alpha
	 */
	public ChatWindow() {
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		setBounds(100, 100, 600, 800);
		setTitle("Envoy");
		setLocationRelativeTo(null);
		setIconImage(Toolkit.getDefaultToolkit().createImage(getClass().getClassLoader().getResource("envoy_logo.png")));

		contentPane.setBorder(new EmptyBorder(space, space, space, space));
		setContentPane(contentPane);
		GridBagLayout gbl_contentPane = new GridBagLayout();
		gbl_contentPane.columnWidths	= new int[] { 1, 1, 1 };
		gbl_contentPane.rowHeights		= new int[] { 1, 1, 1 };
		gbl_contentPane.columnWeights	= new double[] { 0.3, 1.0, 0.1 };
		gbl_contentPane.rowWeights		= new double[] { 0.05, 1.0, 0.07 };
		contentPane.setLayout(gbl_contentPane);

		// TODO: messageList.setFocusTraversalKeysEnabled(false);
		// messageList.setComponentOrientation(ComponentOrientation.LEFT_TO_RIGHT);

		// messageList.setFont(new Font("Arial", Font.PLAIN, 17));
		// messageList.setFixedCellHeight(60);
		messageList.setBorder(new EmptyBorder(space, space, space, space));

		scrollPane.setViewportView(messageList);

		GridBagConstraints gbc_scrollPane = new GridBagConstraints();
		gbc_scrollPane.fill			= GridBagConstraints.BOTH;
		gbc_scrollPane.gridwidth	= 2;
		gbc_scrollPane.gridx		= 1;
		gbc_scrollPane.gridy		= 1;

		gbc_scrollPane.insets = insets;
		contentPane.add(scrollPane, gbc_scrollPane);

		// Message enter field
		messageEnterTextArea.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(KeyEvent e) {
				if (e.getKeyCode() == KeyEvent.VK_ENTER
						&& (Settings.getInstance().isEnterToSend() && e.getModifiersEx() == 0 || e.getModifiersEx() == KeyEvent.CTRL_DOWN_MASK))
					postMessage();
			}
		});

		GridBagConstraints gbc_messageEnterTextfield = new GridBagConstraints();
		gbc_messageEnterTextfield.fill	= GridBagConstraints.BOTH;
		gbc_messageEnterTextfield.gridx	= 1;
		gbc_messageEnterTextfield.gridy	= 2;

		gbc_messageEnterTextfield.insets = insets;

		contentPane.add(messageEnterTextArea, gbc_messageEnterTextfield);

		// Post Button
		GridBagConstraints gbc_moveSelectionPostButton = new GridBagConstraints();

		gbc_moveSelectionPostButton.fill	= GridBagConstraints.BOTH;
		gbc_moveSelectionPostButton.gridx	= 2;
		gbc_moveSelectionPostButton.gridy	= 2;

		gbc_moveSelectionPostButton.insets = insets;

		postButton.addActionListener((evt) -> { postMessage(); });
		contentPane.add(postButton, gbc_moveSelectionPostButton);

		// Settings Button
		GridBagConstraints gbc_moveSelectionSettingsButton = new GridBagConstraints();

		gbc_moveSelectionSettingsButton.fill	= GridBagConstraints.BOTH;
		gbc_moveSelectionSettingsButton.gridx	= 2;
		gbc_moveSelectionSettingsButton.gridy	= 0;

		gbc_moveSelectionSettingsButton.insets = insets;

		settingsButton.addActionListener((evt) -> {
			try {
				new SettingsScreen().setVisible(true);
			} catch (Exception e) {
				logger.log(Level.WARNING, "An error occured while opening the settings screen", e);
				e.printStackTrace();
			}
		});
		contentPane.add(settingsButton, gbc_moveSelectionSettingsButton);

		// Partner name display
		textPane.setFont(new Font("Arial", Font.PLAIN, 20));
		textPane.setEditable(false);

		GridBagConstraints gbc_partnerName = new GridBagConstraints();
		gbc_partnerName.fill	= GridBagConstraints.HORIZONTAL;
		gbc_partnerName.gridx	= 1;
		gbc_partnerName.gridy	= 0;

		gbc_partnerName.insets = insets;
		contentPane.add(textPane, gbc_partnerName);

		userList.setCellRenderer(new UserListRenderer());
		userList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		userList.addListSelectionListener((listSelectionEvent) -> {
			if (client != null && localDb != null && !listSelectionEvent.getValueIsAdjusting()) {
				@SuppressWarnings("unchecked")
				final JList<User>	selectedUserList	= (JList<User>) listSelectionEvent.getSource();
				final User			user				= selectedUserList.getSelectedValue();

				// Select current chat
				currentChat = localDb.getChats().stream().filter(chat -> chat.getRecipient().getId() == user.getId()).findFirst().get();

				// Read current Chat
				try {
					currentChat.read(client);
				} catch (IOException e) {
					e.printStackTrace();
					logger.log(Level.WARNING, "Could notify server about message status change", e);
				}

				// Set chat title
				textPane.setText(currentChat.getRecipient().getName());

				// Update model and scroll down
				messageList.setModel(currentChat.getModel());
				scrollPane.setChatOpened(true);

				messageList.synchronizeModel();
				revalidate();
				repaint();
			}
		});

		userList.setFont(new Font("Arial", Font.PLAIN, 17));
		userList.setBorder(new EmptyBorder(space, space, space, space));

		GridBagConstraints gbc_userList = new GridBagConstraints();
		gbc_userList.fill	= GridBagConstraints.VERTICAL;
		gbc_userList.gridx	= 0;
		gbc_userList.gridy	= 1;
		gbc_userList.anchor	= GridBagConstraints.PAGE_START;
		gbc_userList.insets	= insets;

		applyTheme(Settings.getInstance().getThemes().get(Settings.getInstance().getCurrentTheme()));

		contentPane.add(userList, gbc_userList);
		contentPane.revalidate();

		// Listen to theme changes
		EventBus.getInstance().register(ThemeChangeEvent.class, (evt) -> applyTheme((Theme) evt.get()));

		// Listen to received messages
		EventBus.getInstance().register(MessageCreationEvent.class, (evt) -> {
			Message message = ((MessageCreationEvent) evt).get();
			Chat	chat	= localDb.getChats().stream().filter(c -> c.getRecipient().getId() == message.getSenderId()).findFirst().get();
			chat.appendMessage(message);

			// Read message and update UI if in current chat
			if (chat == currentChat) {
				try {
					currentChat.read(client);
				} catch (IOException e) {
					e.printStackTrace();
					logger.log(Level.WARNING, "Could notify server about message status change", e);
				}
				messageList.synchronizeModel();
			}

			revalidate();
			repaint();
		});

		// Listen to message status changes
		EventBus.getInstance().register(MessageStatusChangeEvent.class, (evt) -> {
			final long			id		= ((MessageStatusChangeEvent) evt).getId();
			final MessageStatus	status	= (MessageStatus) evt.get();

			for (Chat c : localDb.getChats())
				for (Message m : c.getModel())
					if (m.getId() == id) {

						// Update message status
						m.setStatus(status);

						// Update model and scroll down if current chat
						if (c == currentChat) {
							messageList.setModel(currentChat.getModel());
							scrollPane.setChatOpened(true);
						} else messageList.synchronizeModel();
					}

			revalidate();
			repaint();
		});

		revalidate();
	}

	/**
	 * Used to immediately reload the {@link ChatWindow} when settings were changed.
	 *
	 * @param theme the theme to change colors into
	 * @since Envoy v0.2-alpha
	 */
	private void applyTheme(Theme theme) {
		// contentPane
		contentPane.setBackground(theme.getBackgroundColor());
		contentPane.setForeground(theme.getUserNameColor());
		// messageList
		// messageList.setSelectionForeground(theme.getUserNameColor());
		// messageList.setSelectionBackground(theme.getSelectionColor());
		messageList.setForeground(theme.getMessageColorChat());
		messageList.setBackground(theme.getCellColor());
		// scrollPane
		scrollPane.applyTheme(theme);
		scrollPane.autoscroll();

		// messageEnterTextArea
		messageEnterTextArea.setCaretColor(theme.getTypingMessageColor());
		messageEnterTextArea.setForeground(theme.getTypingMessageColor());
		messageEnterTextArea.setBackground(theme.getCellColor());
		// postButton
		postButton.setForeground(theme.getInteractableForegroundColor());
		postButton.setBackground(theme.getInteractableBackgroundColor());
		// settingsButton
		settingsButton.setForeground(theme.getInteractableForegroundColor());
		settingsButton.setBackground(theme.getInteractableBackgroundColor());
		// textPane
		textPane.setBackground(theme.getBackgroundColor());
		textPane.setForeground(theme.getUserNameColor());
		// userList
		userList.setSelectionForeground(theme.getUserNameColor());
		userList.setSelectionBackground(theme.getSelectionColor());
		userList.setForeground(theme.getUserNameColor());
		userList.setBackground(theme.getCellColor());
	}

	private void postMessage() {
		if (userList.isSelectionEmpty()) {
			JOptionPane.showMessageDialog(this, "Please select a recipient!", "Cannot send message", JOptionPane.INFORMATION_MESSAGE);
			return;
		}

		if (!messageEnterTextArea.getText().isEmpty()) try {

			// Create message
			final Message message = new MessageBuilder(localDb.getUser().getId(), currentChat.getRecipient().getId(), localDb.getIdGenerator())
				.setText(messageEnterTextArea.getText())
				.build();

			// Send message
			// TODO: Store offline messages
			client.sendMessage(message);

			// Add message to PersistentLocalDb and update UI
			currentChat.appendMessage(message);

			// Clear text field
			messageEnterTextArea.setText("");

			// Update UI
			revalidate();
			repaint();

			// Request a new id generator if all IDs were used
			if (!localDb.getIdGenerator().hasNext()) client.requestIdGenerator();

		} catch (Exception e) {
			JOptionPane.showMessageDialog(this, "Error sending message:\n" + e.toString(), "Message sending error", JOptionPane.ERROR_MESSAGE);
			e.printStackTrace();
		}
	}

	/**
	 * Initializes the elements of the user list by downloading them from the
	 * server.
	 *
	 * @since Envoy v0.1-alpha
	 */
	private void loadUsersAndChats() {
		new Thread(() -> {
			DefaultListModel<User> userListModel = new DefaultListModel<>();
			localDb.getUsers().values().forEach(user -> {
				userListModel.addElement(user);

				// Check if user exists in local DB
				if (localDb.getChats().stream().filter(c -> c.getRecipient().getId() == user.getId()).count() == 0)
					localDb.getChats().add(new Chat(user));
			});
			SwingUtilities.invokeLater(() -> userList.setModel(userListModel));
		}).start();
	}

	/**
	 * Sets the {@link Client} used by this {@link ChatWindow}.
	 *
	 * @param client the {@link Client} used to send and receive messages
	 * @since Envoy v0.2-alpha
	 */
	public void setClient(Client client) { this.client = client; }

	/**
	 * Sets the {@link LocalDb} used by this {@link ChatWindow}. After
	 * invoking this
	 * method, users and chats will be loaded from the database into the GUI.
	 *
	 * @param localDb the {@link LocalDb} used to manage stored messages
	 *                and users
	 * @since Envoy v0.2-alpha
	 */
	public void setLocalDB(LocalDb localDb) {
		this.localDb = localDb;
		loadUsersAndChats();
	}
}
